// PBUS master 2.0
// Jpsoft (c) 2001, 2009 Jap, http://jap.hu/
// (c) 2008 RCW, walker@omnisterra.com
//
// v1.1 - 20010522 send & receive commands
// v1.2 - 20010522 poll 2-7 device stats
// v1.3 - 20010526 packet reception bug fixed
// v2.0 - 20090816
// v2.2 - 20100221 - correct serial port termios flags
// v2.3 - 20100306 - rewrite in C
// v2.4 - 20100306 - add -s<serial device> option

// this program should be compiled and symlinked
// to cget, cset, cping, cvar, clast, cnoop, etc...

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/time.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/select.h>
#include <limits.h>
#include <errno.h>


#define CGET	0x10
#define CSET	0x11
#define CLAST	0x5b
#define CPING	0x5f
#define CNOOP	0x58
#define CVER	0x5e
#define CRST	0x5c
#define CSTAT	0x5d
#define ROK	0x60
#define RBAD	0x61
#define RECHO	0x6f

typedef unsigned char byte;

struct pkt
{
  byte b0;
  byte cmd;
  byte data[16];
};

void pkt_dump(struct pkt *p)
{
  int dlen = (p->b0)&15;
  int dest = ((p->b0) &240) >>4;
  int cmd = p->cmd;
  int cksum = p->data[dlen];

  printf ("P DEST=%x LEN=%x CMD=%02x SUM=%02x",
   dest, dlen, cmd, cksum);

  if (dlen >0)
  {
    printf (" DATA=");
    int i;
    for (i=0; i<dlen; i++)
      printf("%02x ", p->data[i]);
  }

  printf ("\n");
}

int serial_read(int fd, byte *b)
{
    unsigned int ch = 0;

    int i;
    for (i=0; i<50000; i++)
    {
      int cnt = read (fd, &ch, 1);
      if (cnt > 0)
      {
        *b = ch; printf("R%02x ", ch);
        return (1);
      }
    }

    printf(" -to\n"); return(0); // timeout
}

int pkt_read(int fd, struct pkt *p)
{
/*
      if (!serial_read(fd, &p->b0)) return (0);
      if (!serial_read(fd, &p->cmd)) return (0);
      if (!serial_read(fd, &p->data[0])) return (0);
*/

   int i;
   for (i=0; i<3; i++)
   {
      if (!serial_read(fd, &(p->b0)+i))
         return (0);
   }

  int dlen = (p->b0)&15;
  int dest = ((p->b0) &240) >>4;
  int cmd = p->cmd;

  if (dlen > 0)
    for (i=1 ; i<= dlen; i++)
    {
      if (!serial_read(fd, (byte *)&(p->data)+i))
         return (0);
    }

  int cksum = p->data[dlen];
  return 1;
}

void pkt_write(int fd, struct pkt *p)
{

   unsigned int ch;
   int dlen = (p->b0)&15;
   int cksum = 0;

   int i;
   for (i=0; i<2; i++)
   {
      ch = *((byte *)p+i);
      cksum += ch;
      write (fd, &ch, 1);
      printf("W%02x ", ch);
   }

   if (dlen > 0)
   for (i=0 ; i< dlen; i++)
   {
      ch = p->data[i];
      cksum += ch;
      write (fd, &ch, 1);
      printf("W%02x ", ch);
   }

    cksum = 256 - (cksum&255);
    p->data[dlen] = cksum;

    write (fd, &cksum, 1);
    printf("W%02x\n", cksum);
}

void pkt_fill(struct pkt *p, int cmd, int dest, int index) {
    (*p).b0 = (byte) ((dest << 4) + index);
    (*p).cmd = (byte) cmd;
}

unsigned long getstamp(void)
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return (tv.tv_sec * 1000000 + tv.tv_usec);
}

int main(int argc, char **argv)
{
  struct pkt ps, pd;
//
  char *progname = argv[0];
  int verbose = 1;
  int cmd = CPING;
  int device = 0;
  char *serialdev="/dev/ttyS0";

  extern int optind;
  extern int opterr;
  extern char *optarg;
  int errflag = 0;
  opterr = 0;		/* disables getopt's error msg */

  if (strcmp(progname, "cping") == 0) {
      cmd = CPING;
  } else if (strcmp(progname, "cget") == 0) {
      cmd = CGET;
  } else if (strcmp(progname, "cset") == 0) {
      cmd = CSET;
  } else if (strcmp(progname, "creset") == 0) {
      cmd = CRST;
  } else if (strcmp(progname, "cstat") == 0) {
      cmd = CSTAT;
  } else if (strcmp(progname, "cver") == 0) {
      cmd = CVER;
  } else if (strcmp(progname, "clast") == 0) {
      cmd = CLAST;
  } else if (strcmp(progname, "cnoop") == 0) {
      cmd = CNOOP;
  } else {
      // printf("%s: bad program name\n", progname);
      cmd = CPING;
  }

  if (argc == 1)
      errflag++;  /* print usage if no com args   */

  int c;
  while ((c = getopt(argc, argv, "s:d:c:n:v")) != EOF)
      switch (c) {
      case 's':
          serialdev=optarg;
          break;
      case 'd':
          fflush(stdout);
          device=atoi(optarg);
          break;
      case 'c':
          cmd=atoi(optarg);
          break;
      case 'n':
          // count=atoi(optarg);
          break;
      case 'v':
          verbose++;
          break;
      }
  if (errflag || device==0) {
      fprintf(stderr, "usage: %s [-d<device> -n<count> -c<cmd> -h(help) -s<serial device> -v(verbose)] databytes\n", progname);
      exit(2);
  }

  // collect numerical arguments

  int error=0;
  int index=0;
  char *endptr;

  for (; optind < argc; optind++) {
       char *str = argv[optind];
       errno = 0;
       int val = (int) strtol(str, &endptr, 0);

       if ((errno == ERANGE && (val == LONG_MAX ||
              val == LONG_MIN)) || (errno != 0 && val == 0)) {
           printf("bad numerical argument: %s\n",str);      
           error++;
       }
       if (endptr == str) {
           printf("couldn't convert number: %s\n",str);
           error++;
       }
       if (val < 0 || val > 255) {
           printf("number won't fit in a byte: %s\n",str);
           error++;
       }
       ps.data[index++]=(byte) val;
       // printf("%d\n",val);
  }

  if (index > 15) {
       printf("max payload is 15 bytes\n");
       error++;
  }

  if (error != 0) {
      exit(0);
  }

  char *prog="pbus v2.3";
  int fd=open(serialdev, O_RDWR|O_NONBLOCK);
  if(!fd) {perror(prog);return 1;}

  struct termios t;
  if (tcgetattr(fd, &t) < 0) {perror(prog);return 1;}

  cfsetospeed(&t, B2400);
  cfsetispeed(&t, B2400);
  t.c_lflag = 0; // local modes
  t.c_oflag = 0; // output modes
  t.c_iflag = IGNBRK; // input modes

  if (tcsetattr(fd, TCSANOW, &t)<0) { perror(prog);return 1;}

  // FILE *s = fdopen(fd, "r");
  // setvbuf(s, (char *)NULL, _IONBF, 0);

  pkt_fill(&ps, cmd, device, index);

//
  pkt_write(fd, &ps);
  printf("sent packet: "); pkt_dump(&ps);

  if (pkt_read(fd, &pd))
  {
      printf("rcvd ECHO packet: "); pkt_dump(&pd);
  }
  else
  {
      printf ("ECHO timeout\n");
      exit(1);
  }

  if (pkt_read(fd, &pd))
  {
      printf("rcvd RESP packet: "); pkt_dump(&pd);
  }
else
{
    printf ("RESP timeout\n");
    exit(2);
}

  close(fd);
}
